#define WIN32_LEAN_AND_MEAN 1
#include <windows.h>

LANGUAGE 0, 0

#define VER 3,4,2,0
#define VER_STR "3.4.2\0"

1 VERSIONINFO
FILEVERSION    VER
PRODUCTVERSION VER
FILEFLAGSMASK  VS_FFI_FILEFLAGSMASK
FILEFLAGS      0x00000000L
FILEOS         VOS__WINDOWS32
FILETYPE       VFT_APP
FILESUBTYPE    0x00000000L
BEGIN
	BLOCK "StringFileInfo"
	BEGIN
		BLOCK "040904E4"
		BEGIN
			VALUE "CompanyName",      "Swirc Development Team\0"
			VALUE "FileDescription",  "Curses IRC client\0"
			VALUE "FileVersion",      VER_STR
			VALUE "InternalName",     "swirc\0"
			VALUE "LegalCopyright",   "BSD-3, ISC and MIT\0"
			VALUE "OriginalFilename", "swirc.exe\0"
			VALUE "ProductName",      "Swirc IRC client\0"
			VALUE "ProductVersion",   VER_STR
		END
	END
	BLOCK "VarFileInfo"
	BEGIN
		VALUE "Translation", 0x0409, 1200
	END
END

1 ICON "swirc-royal.ico"
1 RT_MANIFEST "swirc.exe.manifest"
