#ifndef FETCH_DICTIONARY_H
#define FETCH_DICTIONARY_H

__SWIRC_BEGIN_DECLS
extern const char	 g_aff_suffix[];
extern const char	 g_dic_suffix[];

void	cmd_fetchdic(const char *);
__SWIRC_END_DECLS

#endif
