#ifndef SWIRC_EVENT_SERVLIST_H
#define SWIRC_EVENT_SERVLIST_H

__SWIRC_BEGIN_DECLS
void	event_servlist(struct irc_message_compo *) PTR_ARGS_NONNULL;
void	event_servlistEnd(struct irc_message_compo *) PTR_ARGS_NONNULL;
__SWIRC_END_DECLS

#endif
