#ifndef AWAY_H
#define AWAY_H

__SWIRC_BEGIN_DECLS
void	event_away(struct irc_message_compo *);
void	event_unaway(struct irc_message_compo *);
void	event_nowAway(struct irc_message_compo *);
__SWIRC_END_DECLS

#endif
