#ifndef CMDS_MISC_H
#define CMDS_MISC_H

__SWIRC_BEGIN_DECLS
void	cmd_away	(const char *);
void	cmd_banlist	(const char *);
void	cmd_beep	(const char *);
void	cmd_boot	(const char *);
void	cmd_close	(const char *);
void	cmd_cycle	(const char *);
void	cmd_exlist	(const char *);
void	cmd_group	(const char *);
void	cmd_ilist	(const char *);
void	cmd_kill	(const char *);
void	cmd_list	(const char *);
void	cmd_mode	(const char *);
void	cmd_oper	(const char *);
void	cmd_passmod	(const char *);
void	cmd_query	(const char *);
void	cmd_quit	(const char *);
void	cmd_resize	(const char *);
void	cmd_rules	(const char *);
void	cmd_time	(const char *);
void	cmd_version	(const char *);
void	cmd_who		(const char *);
void	cmd_whois	(const char *);
__SWIRC_END_DECLS

#endif
